using AutoScan.Jobs;
using AutoScan.Listener;
using AutoScan.Options;
using Microsoft.Extensions.Logging;
using Microsoft.Extensions.Options;
using Quartz;
using Quartz.Impl.Matchers;

namespace AutoScan;

public class AutoScanApp
{
    private const string GROUP_NAME = "ScanGroup";
    private const string JOBKEY_DOWNLOADER = "downloader";
    private const string JOBKEY_SCANNER = "scanner";
    private const string JOBKEY_CLEANER = "cleaner";
    private readonly AutoScanOptions _options;
    private readonly ILogger<AutoScanApp> _logger;
    private readonly IScheduler _scheduler;
    private readonly IChainerListenerFactory _chainerListenerFactory;
    
    public Func<Task>? OnScanCompleted { get; set; }

    public AutoScanApp(IOptions<AutoScanOptions> options, ILogger<AutoScanApp> logger, IScheduler scheduler, IChainerListenerFactory chainerListenerFactory)
    {
        _options = options.Value;
        _logger = logger;
        _scheduler = scheduler;
        _chainerListenerFactory = chainerListenerFactory;
    }
    
    public async Task Run(CancellationToken cancellationToken)
    {
        if (!_options.Enabled)
        {
            _logger.LogWarning("AutoScanApp is disabled!");
            return;
        }
        _logger.LogInformation("DryRun: {RunDry} | Scan DryRun: {scanRunDry}", _options.RunDry, _options.Scanner?.RunDry);
        _logger.LogInformation("AutoScanApp is running...");

        // var at = DateTime.Now.AddSeconds(10);
        // var cron = $"{at.Second} {at.Minute} {at.Hour} * * ?";
        var cron = CronFromAt(_options.At);
        _logger.LogInformation("Waiting for next scan at {At} [{cron}].", _options.At, cron);
        
        await _scheduler.Start(cancellationToken);
        _logger.LogDebug("Scheduler started successfully!");

        IJobDetail downloaderJob = JobBuilder.Create<DownloaderJob>()
            .WithIdentity(JOBKEY_DOWNLOADER, GROUP_NAME)
            .Build();
        
        //Job with no trigger but tied to the chainer to be executed after the downloader job
        IJobDetail scannerJob = JobBuilder.Create<ScannerJob>()
            .WithIdentity(JOBKEY_SCANNER, GROUP_NAME)
            .StoreDurably()
            .Build();
        
        IJobDetail cleanJob = JobBuilder.Create<CleanJob>()
            .WithIdentity(JOBKEY_CLEANER, GROUP_NAME)
            .StoreDurably()
            .Build();
        
        ITrigger trigger = TriggerBuilder.Create()
            .WithIdentity("trigger1", GROUP_NAME)
            .WithCronSchedule(cron)
            .StartNow()
            .Build();

        var chainer = _chainerListenerFactory.Start("Scan Chainer")
            .AddJob(downloaderJob)
            .AddJob(scannerJob)
            .AddJob(cleanJob)
            .OnFinish(OnScanCompleted)
            .Build();
        
        _scheduler.ListenerManager.AddJobListener(chainer, GroupMatcher<JobKey>.GroupEquals(GROUP_NAME));

        await _scheduler.ScheduleJob(downloaderJob, trigger, cancellationToken);
        await _scheduler.AddJob(scannerJob, false, true, cancellationToken);
        await _scheduler.AddJob(cleanJob, false, true, cancellationToken);
        
        _logger.LogDebug("Scheduled job successfully!");
    }
    
    public async Task StartNewScan()
    {
        var jobKey = new JobKey(JOBKEY_DOWNLOADER, "ScanGroup");
        await _scheduler.TriggerJob(jobKey);
    }

    public string[] GetLastScanPictures()
    {
        if (_options.Scanner?.DetectionFolder is null)
            return [];
        return Directory.GetFiles(_options.Scanner.DetectionFolder , "*.jpg");
    }

    public string? GetVideoPath(string name)
    {
        var path = Path.Combine(_options.Scanner.DetectionFolder, name + ".avi");
        if (!File.Exists(path))
        {
            return null;
        }
        return path;
    }
    
    private string CronFromAt(string at)
    {
        var parts = at.Split(':');
        return $"0 {parts[1]} {parts[0]} * * ?";
    }

}