using AutoScan.Interfaces;
using AutoScan.Models;
using AutoScan.Options;
using Microsoft.Extensions.Logging;
using Microsoft.Extensions.Options;
using System.Net.Http.Json;
using Monitor = AutoScan.Models.Monitor;

namespace AutoScan.Implementations;

public class ShinobiConnector : IDVRConnector
{
    private readonly ILogger<ShinobiConnector> _logger;
    private readonly HttpClient _httpClient;

    private readonly IShinobiLinkFactory _linkFactory;
    private string? _cachedVideoStream;
    
    public ShinobiConnector(ILogger<ShinobiConnector> logger, HttpClient httpClient, IShinobiLinkFactory linkFactory)
    {
        _logger = logger;
        _httpClient = httpClient;
        _linkFactory = linkFactory;
    }

    public async Task<List<VideoDetail>> FetchMonitorVideosBetween(DateTime from, DateTime to, bool runDry = false, CancellationToken cancellationToken = default)
    {
        var endpoint = _linkFactory.BuildVideosLink(from, to);
        
        _logger.LogDebug("Fetching videos details from endpoint: {Endpoint}", endpoint);
        
        if(runDry)
        {
            _logger.LogInformation("RunDry is enabled, skipping video fetch");
            return [];
        }
        
        //get from the server the response with type VideoDetails
        var response = await _httpClient.GetFromJsonAsync<FetchVideoResponse>(endpoint, cancellationToken);
        
        if (response is null)
        {
            _logger.LogWarning("No videos found in the specified range");
            return [];
        }
        
        _logger.LogInformation("Found {Count} videos in the specified range", response.videos.Count);
        
        
        return response.videos.Select(video =>
        {
            video.end = video.end.ToLocalTime();
            video.time = video.time.ToLocalTime();
            _logger.LogDebug("Video: {Filename} - Time: {time} - Ends: {end}", video.filename, video.time, video.end);
            return video;
        }).OrderBy(x => x.time).ToList();
    }

    public async Task DownloadMonitorVideo(VideoDetail video, string downloadFolder, bool runDry, CancellationToken cancellationToken = default)
    {
        var endpoint = _linkFactory.FromHref(video.href);
        _logger.LogDebug("Fetching video from endpoint: {Endpoint}", endpoint);

        //Video filenames format: "monitorId-2025-02-15T07-45-01.mp4"
        var videoTime = video.time.ToString("yyyy-MM-ddTHH-mm-ss");
        var videoPath = Path.Combine(downloadFolder, $"{video.mid}-{videoTime}.{video.ext}");

        try
        {
            _logger.LogDebug("Downloading video...");
            if(runDry)
            {
                _logger.LogInformation("RunDry is enabled, skipping video download");
                return;
            }
            var videoData = await _httpClient.GetByteArrayAsync(endpoint, cancellationToken);
            
            //Write the video to the file
            await File.WriteAllBytesAsync(videoPath, videoData, cancellationToken);
            _logger.LogInformation("Video downloaded successfully!");
        }
        catch (Exception ex)
        {
            _logger.LogError(ex, "An error occurred while downloading the video {video}", video.filename);
            throw;
        }
    }

    
    public async Task<string> GetVideoStream()
    {
        if (_cachedVideoStream is not null)
            return _cachedVideoStream;
        var endpoint = _linkFactory.BuildMonitorLink();
        _logger.LogDebug("Fetching video stream from endpoint: {Endpoint}", endpoint);
        var monitors = await _httpClient.GetFromJsonAsync<Monitor[]>(endpoint);
        
        _cachedVideoStream = monitors?.FirstOrDefault()?.MonitorDetail?.MonitorStreamUrl ?? null;
        return _cachedVideoStream ?? "";
    }
}