using AutoScan.Jobs;
using Microsoft.Extensions.Logging;
using Quartz;
using Quartz.Listener;

namespace AutoScan.Listener;

public class ChainerListener : JobListenerSupport
{
    private readonly ILogger<ChainerListener> _logger;
    private readonly Dictionary<JobKey, JobKey> _chainLinks;
    public Func<Task>? OnJobChainFinished { get; set; }
    
    public ChainerListener(string name, ILogger<ChainerListener> logger)
    {
        _logger = logger;
        Name = name ?? throw new ArgumentException("Listener name cannot be null!");
        _chainLinks = new Dictionary<JobKey, JobKey>();
    }
    public override string Name { get; }
    
    public void AddJobChainLink(JobKey firstJob, JobKey secondJob)
    {
        if (firstJob == null || secondJob == null)
        {
            throw new ArgumentException("Key cannot be null!");
        }
        if (firstJob.Name == null || secondJob.Name == null)
        {
            throw new ArgumentException("Key cannot have a null name!");
        }

        _chainLinks.Add(firstJob, secondJob);
    }

    public override async Task JobWasExecuted(IJobExecutionContext context, JobExecutionException? jobException,
        CancellationToken cancellationToken = default)
    {
        
        if(context.Result is JobResult { Status: JobResultStatus.JobFailed })
        {
            _logger.LogError("There was an error in job {JobKey}. Chain will not continue", context.JobDetail.Key);
            return;
        }

        if (jobException is not null)
        {
            _logger.LogError(jobException, "There was an error in job {JobKey}. Chain will not continue", context.JobDetail.Key);
            return;
        }
        _chainLinks.TryGetValue(context.JobDetail.Key, out var sj);

        if (sj == null)
        {
            //check if it's the last one in the chain
            if (_chainLinks.ContainsValue(context.JobDetail.Key))
            {
                _logger.LogInformation("Job '{JobKey}' is the last in the chain", context.JobDetail.Key);
                if (OnJobChainFinished is not null) 
                    await OnJobChainFinished.Invoke();
                return;
            }
        }

        _logger.LogInformation("Job '{JobKey}' will now chain to Job '{sj}'", context.JobDetail.Key, sj);

        try
        {
            var jobDataMap = new JobDataMap();
            if (context.Result is JobResult { Result: not null } jobResult)
            {
                jobDataMap.Put("previousResult", jobResult.Result);
            }
            await context.Scheduler.TriggerJob(sj, jobDataMap, cancellationToken).ConfigureAwait(false);
        }
        catch (SchedulerException se)
        {
            _logger.LogError(se, "Error encountered during chaining to Job '{sj}'", sj);
        }
    }

}