﻿using AutoScan;
using AutoScan.Options;
using CasaBotApp.Extensions;
using CasaBotApp.TelegramBot;
using ControlServer;
using Microsoft.Extensions.Configuration;
using Microsoft.Extensions.DependencyInjection;
using Microsoft.Extensions.Hosting;
using Microsoft.Extensions.Logging;
using Telegram.Bots;
using Telegram.Bots.Extensions.Polling;
using Telegram.Bots.Http;

var hostBuilder = new HostBuilder();
hostBuilder.ConfigureServices((_, services) =>
{
    var environment = Environment.GetEnvironmentVariable("CASABOT_ENVIRONMENT");
    var configuration = new ConfigurationBuilder()
        .AddJsonFile($"appsettings.json", true, true)
        .AddJsonFile($"appsettings.{environment}.json", true, true)
        .AddEnvironmentVariables()
        .Build();
    
    services.AddSingleton(configuration);
    
    services.AddLogging(configuration);
    
    services.Configure<TelegramOptions>(configuration.GetSection("Telegram"));
    services.Configure<AutoScanOptions>(configuration.GetSection("AutoScan"));
    services.Configure<ShinobiOptions>(configuration.GetSection("Shinobi"));
    services.Configure<ControlServerOptions>(configuration.GetSection("ControlServer"));

    
    services.AddSingleton<IControlServer, ControlServer.ControlServer>();
    services.AddHostedService(sp => sp.GetRequiredService<IControlServer>());
    
    var token = configuration["Telegram:BotToken"] ?? "";
    services.AddSingleton<BotHandler>();
    services.AddBotClient(token);
    services.AddPolling<BotHandler>();
    services.AddSingleton<IUpdateHandler>(sp => sp.GetService<BotHandler>()!);

    services.AddTransient<AlarmBotOrquestrator>();
    
    // To get notifications when a retry is performed
    
    services.AddAutoScan();
    
    services.AddPolicyRegistry().Add("RetryPolicy", RetryPolicyExtension.GetRetryPolicy());
    services.AddHttpClient<IBotClient, BotClient>().AddPolicyHandler(RetryPolicyExtension.GetRetryPolicy());
    
    
    services.Configure<HostOptions>(hostOptions =>
    {
        hostOptions.BackgroundServiceExceptionBehavior = BackgroundServiceExceptionBehavior.Ignore;
        hostOptions.ServicesStartConcurrently = true;
    });

});


var host = hostBuilder.Build();


var logger = host.Services.GetService<ILogger<Program>>()!;
var autoScanApp = host.Services.GetService<AutoScanApp>()!;

var commandRegister = host.Services.GetRequiredService<AlarmBotOrquestrator>();


using var cts = new CancellationTokenSource();

_ = autoScanApp.Run(cts.Token);

commandRegister.RegisterAutoScanApp();
commandRegister.RegisterCommands();
commandRegister.RegisterControlServer();

logger.LogInformation("Bot started");
await host.RunAsync(cts.Token);

await cts.CancelAsync(); // stop the bot




