using Microsoft.Extensions.Logging;
using Microsoft.Extensions.Options;
using System.Text;
using Telegram.Bots;
using Telegram.Bots.Extensions.Polling;
using Telegram.Bots.Requests;
using Telegram.Bots.Types;
using File = System.IO.File;
using SendMediaGroup = Telegram.Bots.Requests.Usernames.SendMediaGroup;
using SendPhotoFile = Telegram.Bots.Requests.Usernames.SendPhotoFile;
using SendText = Telegram.Bots.Requests.Usernames.SendText;
using SendVideoFile = Telegram.Bots.Requests.Usernames.SendVideoFile;

namespace CasaBotApp.TelegramBot;

public class BotHandler : IUpdateHandler
{
    private readonly ILogger<BotHandler> _logger;
    private readonly TelegramOptions _telegramOptions;
    private readonly List<Chat> _subscribers = [];
    private readonly List<Chat> _subscribersAlarm = [];
    private readonly Dictionary<string, BotCommand> _commands;
    //TODO hacerlo mejor.
    private readonly Dictionary<string, CallbackQueueItem> _callbackFunctions = new();
    private record CallbackQueueItem(DateTime inserted, Func<string, long, Task<string>> callback);
    
    public Func<TextMessage, Task>? OnReply { get; set; } = null;

    private readonly IBotClient _bot;

    public BotHandler(IBotClient bot, IOptions<TelegramOptions> telegramConfiguration, ILogger<BotHandler> logger) 
    {
        _logger = logger;
        _telegramOptions = telegramConfiguration.Value;
        _bot = bot;
        _commands = [];
        RegisterCommand(new()
        {
            Command = "/register",
            Description = "Register to receive messages on every update",
            Action = RegisterUser,
            Responder = Respond
        });
        {
            RegisterCommand(new()
            {
                Command = "/registeralarm",
                Description = "Register to receive alarms",
                Action = RegisterUserAlarm,
                Responder = Respond
            });
        }
        RegisterCommand(new()
        {
            Command = "/photo",
            Description = "Get a photo",
            Action = SendImageTest,
            Responder = Respond
        });
        
        foreach (var subs in _telegramOptions.SubscribedChatIds)
        {
            Subscribe(subs);
        }
        foreach(var sub in _telegramOptions.SubscribedAlarmsChatIds)
        {
            SubscribeAlarm(sub);
        }
    }

    public void RegisterCommand(BotCommand command)
    {
        command.Responder = Respond;
        _commands[command.Command] = command;
    }
    
    public void SubscribeAlarm(long id)
    {
        if (_subscribersAlarm.Any(x => x.Id == id))
        {
            _logger.LogWarning("User {Id} is already subscribed to alarm", id);
            return;
        }

        _subscribersAlarm.Add(new Chat()
        {
            Id = id
        });
        _logger.LogInformation("User {Id} subscribed to receive alarms", id);
    }
    
    public void Subscribe(long id)
    {
        if (_subscribers.Any(x => x.Id == id))
        {
            _logger.LogWarning("User {Id} is already subscribed", id);
            return;
        }

        _subscribers.Add(new Chat()
        {
            Id = id
        });
        _logger.LogInformation("User {Id} subscribed to receive messages", id);
    }

    /// <summary>
    /// Send text message to all subscribers
    /// </summary>
    /// <param name="message"></param>
    public Task UpdateText(string message) => UpdateTextInt(_subscribers, message);
    public Task AlertText(string message) => UpdateTextInt(_subscribersAlarm, message);
    public async Task UpdateTextInt(List<Chat> subscribers,string message)
    {
        if (subscribers.Count == 0)
        {
            _logger.LogWarning("No subscribers to send message to");
            return;
        }

        var replacement = new List<(Chat from, Chat to)>();
        foreach (var subscriber in subscribers)
        {
            var response = await SndTxt(subscriber, message);
            if (subscriber.FirstName is null)
            {
                replacement.Add((subscriber, response.Result.Chat));
            }
        }

        foreach (var rep in replacement)
        {
            subscribers.Remove(rep.from);
            subscribers.Add(rep.to);
        }
    }
    

    /// <summary>
    /// Send photo to all subscribers
    /// </summary>
    /// <param name="path"></param>
    /// <param name="caption">Optional message with photo</param>
    /// <param name="options">Actions options</param>
    public Task UpdatePhoto(string path, string? caption = null, IEnumerable<MsgOption>? options = null) =>
        UpdatePhotoInt(_subscribers, path, caption, options);
    
    public Task AlertPhoto(string path, string? caption = null, IEnumerable<MsgOption>? options = null) =>
        UpdatePhotoInt(_subscribersAlarm, path, caption, options);
    
    private async Task UpdatePhotoInt(List<Chat> subscription, string path, string? caption = null, IEnumerable<MsgOption>? options = null)
    {
        if (subscription.Count == 0)
        {
            _logger.LogWarning("No subscribers to send message to");
            return;
        }

        foreach (var subscriber in subscription)
        {
            await using var stream = File.OpenRead(path);
            ReplyMarkup? replyMarkup = null;
            if (options != null)
            {
                foreach(var opt in options)
                {
                    if (opt.type == OptionType.Action && opt.callback != null)
                    {
                        _callbackFunctions.Add(opt.content, new(DateTime.Now, opt.callback));
                    }
                }
                
                // Get options in enumerations of 2
                var optionsGrouped = options.Select((x, i) => (x, i)).GroupBy(x => x.i / 2)
                    .Select(x => x.Select(y => y.x).ToArray()).ToArray();
                IEnumerable<InlineButton[]> keyboard = optionsGrouped.Select(group =>
                    group.Select<MsgOption, InlineButton>(opt => opt.type switch
                    {
                        OptionType.Url => new UrlButton(opt.message, new Uri(opt.content)),
                        OptionType.Action => new CallbackDataButton(opt.message, opt.content),
                        _ => throw new ArgumentOutOfRangeException(nameof(opt), opt, null)
                    }).ToArray()).ToArray();

                replyMarkup = new InlineKeyboardMarkup(keyboard);
            }

            var send = new SendPhotoFile(subscriber.Id.ToString(), stream)
            {
                Caption = caption ?? Path.GetFileNameWithoutExtension(stream.Name),
                ReplyMarkup = replyMarkup
            };
            await _bot.HandleAsync(send);
            _ = HandleQueueExpiration();
        }
    }

    private async Task HandleQueueExpiration()
    {
        //remove expired items with more than 3 hs
        var expired = _callbackFunctions.Where(x => x.Value.inserted.AddHours(3) < DateTime.Now).ToArray();
        foreach (var item in expired)
        {
            _callbackFunctions.Remove(item.Key);
            _logger.LogDebug("Removed expired callback function {Key}", item.Key);
        }
    }

    
    
    /// <summary>
    /// Send photos to all subscribers
    /// </summary>
    /// <param name="paths"></param>
    public async Task UpdatePhotos(string[] paths)
    {
        if (_subscribers.Count == 0)
        {
            _logger.LogWarning("No subscribers to send message to");
            return;
        }
        
        //separate into groups of 10 and send them
        var groups = paths.Select((x, i) => (x, i)).GroupBy(x => x.i / 10).Select(x => x.Select(y => y.x).ToArray()).ToArray();
        foreach (var group in groups)
        {
            await UpdatePhotosInt(group);
        }
    }

    private async Task UpdatePhotosInt(string[] paths)
    {
        var streams = paths.Select(File.OpenRead).ToList();
        var photos = streams.Select(stream => new PhotoFile(stream)
        {
            Caption = Path.GetFileNameWithoutExtension(stream.Name)
        }).Cast<IGroupableMedia>().ToList();
        
        var s0 = _subscribers.FirstOrDefault()?.Id.ToString();
        if(s0 is null)
        {
            _logger.LogWarning("No subscribers to send message to");
            return;
        }
        
        var request = new SendMediaGroup(s0, photos);
        var response = await _bot.HandleAsync(request);
        if (!response.Ok)
        {
            _logger.LogError("Error sending photos.");
        }

        if (_subscribers.Count < 1) return;

        var caches = response.Result.Select(x => (x as PhotoMessage).PhotoSet.LastOrDefault()?.Id).ToList();
        var media = caches.Select(x => new CachedPhoto(x)).ToList();
        
        //send to the rest of the subscribers
        foreach (var subscriber in _subscribers.Skip(1).ToList())
        {
            var request2 = new SendMediaGroup(subscriber.Id.ToString(), media);
            var response2 = await _bot.HandleAsync(request2);
            if (!response2.Ok)
            {
                _logger.LogError("Error sending photos to {Subscriber}", subscriber.Id);
            }
        }
        
        foreach (var stream in streams)
        {
            stream.Close();
            await stream.DisposeAsync();
        }
    }
    
    private async Task SendImageTest(TextMessage msg, BotCommand _)
    {
        await using var stream = File.OpenRead(@"C:\Users\GuillermoMarcel\Pictures\prueba.jpeg");
        
        var send = new SendPhotoFile(msg.Chat.Id.ToString(), stream);
        await _bot.HandleAsync(send);
    }

    public async Task SendPhoto(long chatId, string content, string caption)
    {
        await using var stream = File.OpenRead(content);
        var send = new SendPhotoFile(chatId.ToString(), stream)
        {
            Caption = caption
        };
        var response = await _bot.HandleAsync(send);
        if (!response.Ok)
        {
            _logger.LogError("Error sending photo.");
        }
    }
    
    /// <summary>
    /// Send photos to a specific chat
    /// </summary>
    /// <param name="chatId"></param>
    /// <param name="paths"></param>
    public async Task SendPhotos(long chatId, string[] paths)
    {
        var groups = paths.Select((x, i) => (x, i))
            .GroupBy(x => x.i / 10)
            .Select(x => x.Select(y => y.x).ToArray())
            .ToArray();

        foreach(var group in groups)
        {
            await SendPhotosInt(chatId, group);
        }
    }

    private async Task SendPhotosInt(long chatId, string[] paths)
    {
        var streams = paths.Select(File.OpenRead).ToArray();
        var photos = streams.Select(stream => new PhotoFile(stream)
        {
            Caption = Path.GetFileNameWithoutExtension(stream.Name)
        }).Cast<IGroupableMedia>().ToArray();
        
        var request = new SendMediaGroup(chatId.ToString(), photos);
        var response = await _bot.HandleAsync(request);
        if (!response.Ok)
        {
            _logger.LogError("Error sending photos.");
        }

        foreach (var stream in streams)
        {
            stream.Close();
            await stream.DisposeAsync();
        }
    }

    /// <summary>
    /// Send Text Message to a specific chat
    /// </summary>
    /// <param name="chatId"></param>
    /// <param name="textMessage"></param>
    public async Task SendText(long chatId, string textMessage)
    {
        var response = await SndTxt(chatId, textMessage);
        if (!response.Ok)
        {
            _logger.LogError("Error sending message to {ChatId}", chatId);
        }
    }
    
    /// <summary>
    /// Send single video to a specific chat
    /// </summary>
    /// <param name="chatId"></param>
    /// <param name="path"></param>
    public async Task SendVideo(long chatId, string path)
    {
        _logger.LogInformation("Sending video to {ChatId}", chatId);
        await using var stream = File.OpenRead(path);
        var send = new SendVideoFile(chatId.ToString(), stream);
        var response = await _bot.HandleAsync(send);
        if (!response.Ok)
        {
            _logger.LogError("Error sending video.");
        }
        _logger.LogInformation("Video sent to {ChatId}", chatId);
        stream.Close();
    }


    private async Task OnMessage(TextMessage msg)
    {
        if(!_commands.TryGetValue(msg.Text, out var command))
        {
            if (msg.ReplyToMessage != null && OnReply is not null)
            {
                await OnReply(msg);
                return;
            }
            await SendHelp(msg);
            return;
        }

        if (command?.Action is null)
        {
            _logger.LogError("Command {Command} has no action", msg.Text);
        }
        
        await command!.Action!(msg, command);
    }

    private async Task SendHelp(TextMessage msg)
    {
        _logger.LogInformation("Received '{Text}' in {Chat}", msg.Text, msg.Chat);
        var message = new StringBuilder();
        message.AppendLine("Commands:");
        foreach (var command in _commands.Values)
        {
            message.AppendLine($"{command.Command} - {command.Description}");
        }
        await Respond(msg, message.ToString());
    }

    private async Task RegisterUser(TextMessage msg, BotCommand _)
    {
        if (_subscribers.Any(c => c.Id == msg.Chat.Id))
        {
            await Respond(msg, "You are already registered to receive messages");
            return;
        }

        _subscribers.Add(msg.Chat);
        _logger.LogInformation("User {User} ({id}) registered to receive messages", msg.Chat.FirstName, msg.Chat.Id);
        await Respond(msg, "You are registered to receive messages every minute");
    }
    private async Task RegisterUserAlarm(TextMessage msg, BotCommand _)
    {
        if (_subscribersAlarm.Any(c => c.Id == msg.Chat.Id))
        {
            await Respond(msg, "You are already registered to receive Alarms");
            return;
        }

        _subscribersAlarm.Add(msg.Chat);
        _logger.LogInformation("User {User} ({id}) registered to receive Alarms", msg.Chat.FirstName, msg.Chat.Id);
        await Respond(msg, "You are registered to receive alarms");
    }

    
    private Task<Response<TextMessage>> SndTxt(long id, string txt) => _bot.HandleAsync(new SendText(id.ToString(), txt));
    private Task<Response<TextMessage>> SndTxt(Chat chat, string txt) => SndTxt(chat.Id, txt);
    private Task<Response<TextMessage>> Respond(Message msg, string txt) => SndTxt(msg.Chat.Id, txt);
    private Task<Response<PhotoMessage>> SndPhoto(Chat chat, FileStream path) => 
        _bot.HandleAsync(new SendPhotoFile(chat.Id.ToString(), path));


    public Task HandleAsync(IBotClient bot, Update update, CancellationToken cst)
    {
        try
        {
            return update switch
            {
                MessageUpdate u when u.Data is TextMessage message =>
                    OnMessage(message),

                EditedMessageUpdate u when u.Data is TextMessage message =>
                    bot.HandleAsync(new SendText(message.Chat.Id.ToString(), message.Text)
                    {
                        ReplyToMessageId = message.Id
                    }, cst),
                
                CallbackQueryUpdate u when u.Data is MessageCallbackQuery resp =>
                    HandleCallbackResponse(u, resp),
                    
                _ => Task.CompletedTask
            };
        }catch (Exception e)
        {
            _logger.LogError(e, "Error handling update");
            return Task.CompletedTask;
        }   
    }

    private async Task HandleCallbackResponse(CallbackQueryUpdate update, MessageCallbackQuery response)
    {
        _logger.LogInformation("Callback message received: {id} {Response}", update.Id, response.Data);
        if (!_callbackFunctions.TryGetValue(response.Data, out var queueCallback))
        {
            _logger.LogError("Callback function not found for {Response}", response.Data);
            _ = _bot.HandleAsync(new AnswerCallbackQuery(response.Id, "Option expired"));
            return;
        }
        var chatId = response.Message.Chat.Id;
        
        var answer = await queueCallback.callback(response.Data, chatId);

        var r = await _bot.HandleAsync(new AnswerCallbackQuery(response.Id, answer));
        if (!r.Ok)
        {
            _logger.LogError("Error sending callback response: {Error}", r.Failure);
        }
    }


    
}