﻿using Microsoft.Extensions.Logging;
using Microsoft.Extensions.Options;
using System.Net;
using System.Text;

namespace ControlServer;

public class ControlServer : IControlServer
{
    private readonly ILogger<ControlServer> _logger;
    private readonly ControlServerOptions _options;
    private readonly HttpListener _httpListener;
    private readonly Dictionary<long, SensorEvent> _events = new();
    private Func<SensorEvent, Task>? _onEventRecived;
    private bool _disarmRequestPending;

    public ControlServer(ILogger<ControlServer> logger, IOptions<ControlServerOptions> options)
    {
        _logger = logger;
        _options = options.Value;
        _httpListener = new HttpListener();
        _httpListener.Prefixes.Add($"http://+:{_options.port}/");
    }

    public async Task StartAsync(CancellationToken ct)
    {
        try
        {
            _httpListener.Start();
            _logger.LogInformation("Http server listening on port {Port}", _options.port);
        }catch (Exception e)
        {
            _logger.LogError("Error starting HTTP server: {Message}", e.Message);
            return;
        }

        while (!ct.IsCancellationRequested)
        {
            var ctx = await _httpListener.GetContextAsync();
            // _logger.LogDebug("Got a request");
            
            var request = ctx.Request;
            var response = ctx.Response;
            
            // _logger.LogInformation("Url: {Url} | Method {method}", request.Url, request.HttpMethod);
            
            //Query params
            var eId = request.QueryString["eventId"];
            var eventType = request.QueryString["eventType"];
            var eventData = request.QueryString["eventData"];
            var eventId = long.TryParse(eId, out var id) ? id : 0;
            var eventTypeEnum = Enum.TryParse<EventType>(eventType, out var type) ? type : EventType.Unknown;
            var eventDataString = eventData ?? string.Empty;
            var sensorEvent = new SensorEvent(eventId, eventTypeEnum, eventDataString);
            

            if (sensorEvent.Type == EventType.Update)
            {
                HandleUpdate(response);
                response.Close();
                continue;
            }
            _logger.LogInformation("Event Received: {event}", sensorEvent);

            var notify = HandleSensorEvent(sensorEvent);
            
            
            response.StatusCode = 200;
            response.ContentType = "text/plain";
            
            var message = sensorEvent.Authorization == Authorization.Unauthorized ? "UNAUTHORIZED" : "AUTHORIZE_ENTRANCE";
            response.ContentLength64 = Encoding.UTF8.GetByteCount(message);
            await using (var writer = new StreamWriter(response.OutputStream))
            {
                await writer.WriteAsync(message);
            }
            response.Close();
            
            if (notify)
            {
                _ = _onEventRecived?.Invoke(sensorEvent);
            }
        }
        
    }

    private readonly List<Card> _cards = new()
    {
        new Card("1", "Card 1"),
        new Card("2", "Card 2"),
        new Card("3", "Card 3"),
        new Card("4", "Card 4"),
        new Card("5", "Card 5"),
    };
    private void HandleUpdate(HttpListenerResponse response)
    {
        var updateResponse = new UpdateResponse(_disarmRequestPending, _cards.ToArray());
        _disarmRequestPending = false;
        
        response.StatusCode = 200;
        response.ContentType = "application/json";
        // response.ContentType = "text/plain";
        var message = System.Text.Json.JsonSerializer.Serialize(updateResponse);
        response.ContentLength64 = Encoding.UTF8.GetByteCount(message);
        using var writer = new StreamWriter(response.OutputStream);
        writer.Write(message);
    }

    public record UpdateResponse(bool disarm, Card[] cards);
    public record Card(string id, string name);
    private bool HandleSensorEvent(SensorEvent se)
    {
        if (se.Type == EventType.Update && _disarmRequestPending)
        {
            _disarmRequestPending = false;
            se.Authorization = Authorization.Authorized;
            return false;
        }
        
        _events.TryGetValue(se.EventId, out var storedEvent);
        if (storedEvent is null)
        {
            //New One
            if (se.Type == EventType.DisarmedEntrance)
            {
                _disarmRequestPending = false;
                se.Authorization = Authorization.Authorized;
                _events.Add(se.EventId, se);
                return true;
            }

            if (se.Type == EventType.Fired)
            {
                //Check pending desarmed.
                se.Authorization = _disarmRequestPending ? Authorization.Authorized : Authorization.Unauthorized;
                _disarmRequestPending = false;
                _events.Add(se.EventId, se);
                return true;
            }
            _events.Add(se.EventId, new SensorEvent(se.EventId, se.Type, se.Data));
            return false;
        }
        
        //Already stored
        if (se.Type == EventType.EventUpdate && (_disarmRequestPending || storedEvent.Authorization == Authorization.Authorized) )
        {
            _disarmRequestPending = false;
            storedEvent.Authorization= Authorization.Authorized;
            se.Authorization = Authorization.Authorized;
            return false;
        }        
        return false;
    }

    public Task StopAsync(CancellationToken ct)
    {
        // _httpListener.EndGetContext(null);
        // _httpListener.Close();
        _httpListener.Stop();
        return Task.CompletedTask;
    }

    public void OnEvent(Func<SensorEvent, Task> onEventRecived)
    {
        _onEventRecived = onEventRecived;
    }

    public void AuthorizeEvent(long eventId)
    {
        if (_events.TryGetValue(eventId, out var sensorEvent))
        {
            sensorEvent.Authorization = Authorization.Authorized;
        }
    }

    public void RequestDisarm()
    {
        _disarmRequestPending = true;
    }

}